

#include "vgev.h"

#ifdef WIN32

#include <GL/gl.h>
#include <time.h>

#else
#include <X11/keysym.h>
#include <string.h>
#endif

#include "timer.h"
#include "global.h"

#include "vgev_data.h"


#define MARGIN 5.0

#ifdef WIN32

extern iface *isim;


/* ****************************************************************
**************************************************************** */
LRESULT CALLBACK windrawproc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {

   event_message_type event;
   canvas *ptr;

   event.window = hwnd;
   event.type   = message;
   event.wparam = wParam;
   event.lparam = lParam;

   for (ptr = (canvas *)canvas_manager.head; ptr && ptr->win.mwindow != hwnd; ptr = (canvas *)ptr->next);

   if (ptr)
      ptr->process_input(isim, &event);

   return DefWindowProc(hwnd, message, wParam, lParam);
}


#endif


/* ****************************************************************
**************************************************************** */
canvas::canvas(int mc) {

   int i;
   double t, dt;

   xoffset = 0;
   yoffset = 0;
   xcenter = WINX>>1;
   ycenter = WINY>>1;

   scale = 1.0;

   mode_count = mc;
   
   if (mode_count)
      mode_stat = new mode_status[mode_count];
   else
      mode_stat = NULL;
   
   mouse_mode = new mouse_handler_type(this);
   flags = FLAG_UPDATE;
   redraw_flag = -1;

   t = 0;
   dt = TWOPI/31.0;

   for (i=0; i<31; i++, t+=dt) {
      circle[i][0] = cos(t);
      circle[i][1] = sin(t);
   }

   circle[i][0] = 1;
   circle[i][1] = 0;
}
      
      
/* ****************************************************************
**************************************************************** */
canvas::canvas(int mc, mouse_handler_type *m) {

   xoffset = 0;
   yoffset = 0;
   xcenter = WINX>>1;
   ycenter = WINY>>1;

   scale = 1.0;

   mode_count = mc;
   
   if (mode_count)
      mode_stat = new mode_status[mode_count];
   else
      mode_stat = NULL;
      
   mouse_mode = m;
   mouse_mode->set_canvas(this);
   
   flags = FLAG_UPDATE;
   redraw_flag = -1;
}
      
      
/* ****************************************************************
**************************************************************** */
void canvas::tpColor3ub(unsigned char r, unsigned char g, unsigned char b, int gflag) {

   int i;
   
   if (flags & FLAG_GREYSCALE) {
      if (gflag) {
         glColor3ub(0, 64, 0);
         return;
      }
      
      i = r;
      if (i < g)
         i = g;
      if (i < b)
         i = b;
      
      glColor3ub(i, i, i);
      return;
   }
   
   glColor3ub(r, g, b);
}


/* ****************************************************************
**************************************************************** */
void canvas::tpColor4ubv(unsigned char *v, int gflag) {

   int i;
   
   if (flags & FLAG_GREYSCALE) {
      if (gflag) {
         glColor4ub(0, 64, 0, v[3]);
         return;
      }

      i = v[0];
   
      if (i < v[1])
         i = v[1];
      if (i < v[2])
         i = v[2];
      
      glColor4ub(i, i, i, v[3]);
      return;
   }
   
   glColor4ubv(v);
}


/* ****************************************************************
**************************************************************** */
void canvas::tpColor3ubv(unsigned char *v, int gflag) {

   int i;
   
   if (flags & FLAG_GREYSCALE) {
      if (gflag) {
         glColor3ub(0, 64, 0);
         return;
      }

      i = v[0];

      if (i < v[1])
         i = v[1];
      if (i < v[2])
         i = v[2];
      
      glColor3ub(i, i, i);
      return;
   }
   
   glColor3ubv(v);
}


/* ****************************************************************
**************************************************************** */
void canvas::init(int argc, char **argv) {

   dbl_llist_manager man;
   int i;
       
   for (i=0; i<argc; i++) {
      if (!strcmp(argv[i], "-greyscale")) {
         flags |= FLAG_GREYSCALE;
         continue;
      }

   }
      
#ifdef WIN32

   unsigned int attribs[] = { OGLFLAG_RGBA, OGLFLAG_DOUBLEBUFFER, OGLFLAG_BPP, 16, OGLFLAG_NULL };
   RECT rect;

   GetWindowRect(GetDesktopWindow(), &rect);
   WINX = rect.right - rect.left;
   WINY = rect.bottom - rect.top;

   win.VIRTUALcancel();
   win.VIRTUALreset(argc, argv, WINX, WINY, "GEV", attribs);

   ShowCursor(1);

   SetWindowLong(win.mwindow, GWL_WNDPROC, (long)windrawproc);

#else

   unsigned int attribs[] = { OGLFLAG_RGBA, OGLFLAG_DOUBLEBUFFER, OGLFLAG_NULL };
   Display *mdisplay;
   XWindowAttributes wa;

   mdisplay = XOpenDisplay(NULL);
   XGetWindowAttributes(mdisplay, XRootWindow (mdisplay, XDefaultScreen(mdisplay)), &wa);
   XCloseDisplay(mdisplay);

   WINX = wa.width;
   WINY = wa.height;

   win.VIRTUALcancel();
   win.VIRTUALreset(argc, argv, WINX, WINY, "GEV", attribs);

   XSelectInput(win.mdisplay, win.mwindow, ButtonPressMask | ButtonReleaseMask |
                PointerMotionMask | KeyReleaseMask | ExposureMask | StructureNotifyMask);
  
//   XDefineCursor(win.mdisplay,win.mwindow,cursor);

#endif

   tileman.dest();
   primitive_manager.dest();
   
   if (map_indexname.string)
      tileman.read_tiles(map_indexname.string);
}


/* ****************************************************************
**************************************************************** */
void canvas::resize(int w, int h) {

   double s;

   s = (scale *w)/(double)WINX;
   set_scale(&s);

   WINX = w;
   WINY = h;
   
   xcenter = WINX>>1;
   ycenter = WINY>>1;

   symbol_manager.flags |= SYMBOLFLAG_RECALC;
   flags |= FLAG_UPDATE;
   redraw_flag = -1;
}


/* *************************************************************************************
************************************************************************************* */
void canvas::recenter(double *x, double *y) {

   xoffset += -(*x - xcenter) / scale; 
   yoffset += -(*y - ycenter) / scale;

   symbol_manager.flags |= SYMBOLFLAG_RECALC;
   flags |= FLAG_UPDATE;
   redraw_flag = -1;
}


/* ****************************************************************
**************************************************************** */
void canvas::xy2lonlat(int x, int y, double *lon, double *lat) {

   *lon = (x - xcenter) / scale - xoffset;
   *lat = (y - ycenter) / scale - yoffset;
}


/* ****************************************************************
**************************************************************** */
int canvas::clip(double lon, double lat, int *xout, int *yout) {

   int clipcode;

   *xout = (int)((lon + xoffset) * scale + xcenter);
   *yout = (int)((lat + yoffset) * scale + ycenter);

   clipcode =  *xout > ((int)(WINX-1)) ? 1 : 0;
   clipcode |= *xout < 0 ? 2 : 0;
   clipcode |= *yout > ((int)(WINY-1)) ? 4 : 0;
   clipcode |= *yout < 0 ? 8 : 0;

   return clipcode;
}


/* ****************************************************************
**************************************************************** */
void canvas::recenter(double *radius) {

   double lat, lon;
   
   win.VIRTUALgetdim((int *)&WINX, (int *)&WINY);

   xoffset = -rad2deg(CONVERT_ctrlon);
   yoffset = -rad2deg(CONVERT_ctrlat);

   scenario_radius = *radius;

   xy_to_deg(scenario_radius,scenario_radius, &lat, &lon);

   lon = WINX/(2.0*(lon+xoffset));
   set_scale(&lon);

   xcenter = WINX>>1;
   ycenter = WINY>>1;

   symbol_manager.flags |= SYMBOLFLAG_RECALC;
   flags |= FLAG_UPDATE;
   redraw_flag = -1;
}


/* ****************************************************************
**************************************************************** */
void canvas::set_scale(double *s) {

   scale = *s;

   // set limit to zone out at max 360 degrees
   if (WINX/(360.0+MARGIN) > scale)
      scale = WINX/(360.0+MARGIN);

   // set limit to zone in to 1 second of a degree
   else if (WINX*3600 < scale)
      scale = WINX*3600;

   symbol_manager.flags |= SYMBOLFLAG_RECALC;
   flags |= FLAG_UPDATE;
   redraw_flag = -1;
}


#ifdef WIN32


/* ****************************************************************
**************************************************************** */
int canvas::process_input(iface *isim, void *event) {

   double x, y;
   event_message_type *mevent = (event_message_type *)event;
   int w, h;

   switch (mevent->type) {

      case WM_SIZE:
         win.VIRTUALgetdim(&w, &h);
      
         if (w != (int)WINX || h != (int)WINY)
            resize(w, h);

         return 1;

      case WM_MOUSEMOVE:
         x =  ((double)LOWORD(mevent->lparam) - xcenter) / scale - xoffset;
         y = -((double)HIWORD(mevent->lparam) - ycenter) / scale - yoffset;

         mouse_mode->event_motion(LOWORD(mevent->lparam), WINY-HIWORD(mevent->lparam), x, y);
         redraw_flag = 13;
         return 1;

      case WM_LBUTTONUP:
         x =  ((double)LOWORD(mevent->lparam) - xcenter) / scale - xoffset;
         y = -((double)HIWORD(mevent->lparam) - ycenter) / scale - yoffset;

         mouse_mode->event_release(isim, LEFTBUTTON, LOWORD(mevent->lparam), WINY-HIWORD(mevent->lparam), x, y);
         redraw_flag = 14;
         return 1;

     case WM_MBUTTONUP:
         x =  ((double)LOWORD(mevent->lparam) - xcenter) / scale - xoffset;
         y = -((double)HIWORD(mevent->lparam) - ycenter) / scale - yoffset;

         mouse_mode->event_release(isim, MIDDLEBUTTON, LOWORD(mevent->lparam), WINY-HIWORD(mevent->lparam), x, y);
         redraw_flag = 14;
         return 1;

     case WM_RBUTTONUP:
         x =  ((double)LOWORD(mevent->lparam) - xcenter) / scale - xoffset;
         y = -((double)HIWORD(mevent->lparam) - ycenter) / scale - yoffset;

         mouse_mode->event_release(isim, RIGHTBUTTON, LOWORD(mevent->lparam), WINY-HIWORD(mevent->lparam), x, y);
         redraw_flag = 14;
         return 1;

      case WM_LBUTTONDOWN:
         x =  ((double)LOWORD(mevent->lparam) - xcenter) / scale - xoffset;
         y = -((double)HIWORD(mevent->lparam) - ycenter) / scale - yoffset;

         mouse_mode->event_pick(LEFTBUTTON, LOWORD(mevent->lparam), WINY-HIWORD(mevent->lparam), x, y);
         redraw_flag = 15;
         return 1;
       
      case WM_MBUTTONDOWN:
         x =  ((double)LOWORD(mevent->lparam) - xcenter) / scale - xoffset;
         y = -((double)HIWORD(mevent->lparam) - ycenter) / scale - yoffset;

         mouse_mode->event_pick(MIDDLEBUTTON, LOWORD(mevent->lparam), WINY-HIWORD(mevent->lparam), x, y);
         redraw_flag = 15;
         return 1;

      case WM_RBUTTONDOWN:
         x =  ((double)LOWORD(mevent->lparam) - xcenter) / scale - xoffset;
         y = -((double)HIWORD(mevent->lparam) - ycenter) / scale - yoffset;

         mouse_mode->event_pick(RIGHTBUTTON, LOWORD(mevent->lparam), WINY-HIWORD(mevent->lparam), x, y);
         redraw_flag = 15;
         return 1;

      case WM_PAINT:
         redraw_flag = 16;
         return 1;

      case WM_KEYDOWN:
         if (VK_ESCAPE == (mevent->wparam & 0xff))
            flags |= FLAG_DONE;
    
         return 1;

      default:
         return 0;
   }
      
}

#else

/* ****************************************************************
**************************************************************** */
int canvas::process_input(iface *isim, void *event) {

   double x, y;
   int w, h;

   switch (((XEvent *)event)->type) {

      case ConfigureNotify:
         win.VIRTUALgetdim(&w, &h);
      
         if (w != (int)WINX || h != (int)WINY)
            resize(w, h);
 
         return 1;
 
      case MotionNotify:
         x =  (((XEvent *)event)->xmotion.x - xcenter) / scale - xoffset;
         y = -(((XEvent *)event)->xmotion.y - ycenter) / scale - yoffset;

         mouse_mode->event_motion(((XEvent *)event)->xmotion.x, WINY-((XEvent *)event)->xmotion.y, x, y);
         redraw_flag = 18;
         return 1;

      case ButtonRelease:
         switch (((XEvent *)event)->xbutton.button) {
            case Button1:
               x =  (((XEvent *)event)->xmotion.x - xcenter) / scale - xoffset;
               y = -(((XEvent *)event)->xmotion.y - ycenter) / scale - yoffset;

               mouse_mode->event_release(isim, LEFTBUTTON, ((XEvent *)event)->xbutton.x, WINY-((XEvent *)event)->xbutton.y, x, y);
               redraw_flag = 19;
               break;
       
            case Button2: // middle
               x =  (((XEvent *)event)->xmotion.x - xcenter) / scale - xoffset;
               y = -(((XEvent *)event)->xmotion.y - ycenter) / scale - yoffset;

               mouse_mode->event_release(isim, MIDDLEBUTTON, ((XEvent *)event)->xbutton.x, WINY-((XEvent *)event)->xbutton.y, x, y);
               redraw_flag = 19;
               break;
       
            case Button3: // right
               x =  (((XEvent *)event)->xmotion.x - xcenter) / scale - xoffset;
               y = -(((XEvent *)event)->xmotion.y - ycenter) / scale - yoffset;

               mouse_mode->event_release(isim, RIGHTBUTTON, ((XEvent *)event)->xbutton.x, WINY-((XEvent *)event)->xbutton.y, x, y);
               redraw_flag = 19;
               break;
       
            default:
               break;       
         }

         return 1;

      case ButtonPress:
         switch (((XEvent *)event)->xbutton.button) {
            case Button1:
               x =  (((XEvent *)event)->xmotion.x - xcenter) / scale - xoffset;
               y = -(((XEvent *)event)->xmotion.y - ycenter) / scale - yoffset;

               mouse_mode->event_pick(LEFTBUTTON, ((XEvent *)event)->xbutton.x, WINY-((XEvent *)event)->xbutton.y, x, y);
               redraw_flag = 17;
               break;
       
            case Button2:   // middle
               x =  (((XEvent *)event)->xmotion.x - xcenter) / scale - xoffset;
               y = -(((XEvent *)event)->xmotion.y - ycenter) / scale - yoffset;

               mouse_mode->event_pick(MIDDLEBUTTON, ((XEvent *)event)->xbutton.x, WINY-((XEvent *)event)->xbutton.y, x, y);
               redraw_flag = 17;
               break;
           
            case Button3:   // right
               x =  (((XEvent *)event)->xmotion.x - xcenter) / scale - xoffset;
               y = -(((XEvent *)event)->xmotion.y - ycenter) / scale - yoffset;

               mouse_mode->event_pick(RIGHTBUTTON, ((XEvent *)event)->xbutton.x, WINY-((XEvent *)event)->xbutton.y, x, y);
               redraw_flag = 17;
               break;
       
            default:
               break;       
         }

         return 1;

      case Expose:
         redraw_flag = 20;
         return 1;

      case KeyRelease:
         switch (win.Xcode2sym(((XEvent *)event)->xkey.keycode)) {
            case XK_Escape:
               flags |= FLAG_DONE;
               break;

            case XK_space:
               if (flags & FLAG_FOLLOW)
                  flags &= ~FLAG_FOLLOW;
               else
                  flags |= FLAG_FOLLOW;

               break;

            default:
               break;
         }
 
         return 1;

      default:
         return 0;
   }
      
}

#endif


/* ****************************************************************
**************************************************************** */
void canvas::update_display_background() {

   int x, y, w, h, j;
   double dx, dy;
   double lat, lon;

   glViewport(0, 0, WINX, WINY);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(0,WINX,0,WINY, -1, 1);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glRasterPos2i(0,0);

   // Draw map and overlays with current range scale and game area center 
   if ((flags & (FLAG_UPDATE | FLAG_FOLLOW))) {
      earth_map.preprocess_map(this);
      tileman.preprocess_map(this);
   }

   glClear(GL_COLOR_BUFFER_BIT);

   earth_map.render(this);   
   tileman.render(this);

   // scenario limits
   if (scenario_radius > CORRECT) {

      tpColor3ub(255, 0, 0, 1);

      if (!clip(rad2deg(CONVERT_ctrlon),rad2deg(CONVERT_ctrlat), &x, &y)) {
    
         glBegin(GL_LINES);
            glVertex2i(x-5, y);
            glVertex2i(x+5, y);
            glVertex2i(x, y-5);
            glVertex2i(x, y+5);
         glEnd();
      }

      glBegin(GL_LINE_LOOP);

         for (j=0; j<360; j+=5) {
            dx = cos(deg2rad((double)j))*scenario_radius;
            dy = sin(deg2rad((double)j))*scenario_radius;
            xy_to_deg(dx, dy, &lat, &lon);
            clip(lon, lat, &w, &h);
            glVertex2i(w, h);
         }
   
      glEnd();
   }

   flags &= ~FLAG_UPDATE;
}


/* ****************************************************************
**************************************************************** */
void canvas::update_display() {

   static double timer = 128;
   static double cyclic_timer = 0;
   static unsigned int toggle = 0;
   static clock_t oldtime = clock();

   long         time_in_secs;
   char         buffer[MAXSTRLEN];
   clock_t      newtime;
   drawable_symbol *dstr;
   int i;
 
   if (!redraw_flag)
      return;
      
   update_display_background();

   // crosshairs
   tpColor3ub(200, 200, 255, 1);

   glBegin(GL_LINES);
      glVertex2i(mouse_mode->mouse_pos[0]-5, mouse_mode->mouse_pos[1]);
      glVertex2i(mouse_mode->mouse_pos[0]+5, mouse_mode->mouse_pos[1]);
      glVertex2i(mouse_mode->mouse_pos[0], mouse_mode->mouse_pos[1]-5);
      glVertex2i(mouse_mode->mouse_pos[0], mouse_mode->mouse_pos[1]+5);
   glEnd();

   // Draw text
   
   time(&time_in_secs);
#ifndef SGI
   sprintf(buffer, "%s", asctime(gmtime(&time_in_secs)));
   buffer[strlen(buffer)-1] = 0;
   strcat(buffer, " ZULU");
#else
   ascftime(buffer,"DATE: %d %b %y  TIME: %H:%M:%S ZULU",gmtime(&time_in_secs));
#endif

   tpColor3ub(205, 205, 0, 0);

   DrawString((WINX>>1) - strlen(buffer) * 4,WINY-20,buffer,strlen(buffer));

   sprintf(buffer,"RANGE SCALE: %d NM",(int)((deg2rad(WINX * CONVERT_EarthRadius * MET2NMI))/scale));
   DrawString((WINX>>1) - strlen(buffer) * 3,WINY-40,buffer,strlen(buffer));

   // Draw classification title
   tpColor3ubv(classcolor, 0);

   DrawString(20,WINY-20,classification,strlen(classification));
   DrawString(20,10,classification,strlen(classification));

   // write lat/lon to screen
   tpColor3ub(205, 205, 0, 0);
   DrawString(WINX-175,WINY-15,readout.string,readout.stringlen());

   // mouse mode
   switch (mouse_mode->query_state()) {

      // draw zoom box
      case VGEV_MOUSE_MODE_ZOOM:
         DrawString(WINX-175,WINY-30,"* Box Zoom *", 12);
         break;

      case VGEV_MOUSE_MODE_AZRANGE:
         DrawString(WINX-175,WINY-30,"* Az-Range *", 12);
         break;

      case VGEV_MOUSE_MODE_PICK:
         DrawString(WINX-175,WINY-30,"* Pick *", 8);
         break;

      case VGEV_MOUSE_MODE_RECENTER:
         DrawString(WINX-175,WINY-30,"* Recenter *", 12);
         break;

      case VGEV_MOUSE_MODE_LOS:
         DrawString(WINX-175,WINY-30,"* Line of Sight *", 17);
         break;

      default:
         break;
   }

   if (symbol_manager.flags & SYMBOLFLAG_RECALC)
      symbol_manager.recalc(this);
      
   // render symbols
   for (dstr=(drawable_symbol *)symbol_manager.head; dstr; dstr=(drawable_symbol *)dstr->next)
      dstr->render(this);

   // misc graphics
   if (mouse_mode->drawflag)
      switch (mouse_mode->query_state()) {

         // draw zoom box
         case VGEV_MOUSE_MODE_ZOOM:
            if (!mouse_mode->enable)
               break;

            tpColor3ub((unsigned char)timer, (unsigned char)timer, (unsigned char)timer, 1);

            glBegin(GL_LINE_LOOP);
               glVertex2iv(mouse_mode->box[0]);
               glVertex2i(mouse_mode->box[1][0], mouse_mode->box[0][1]);
               glVertex2iv(mouse_mode->box[1]);
               glVertex2i(mouse_mode->box[0][0], mouse_mode->box[1][1]);
            glEnd();

            break;

         case VGEV_MOUSE_MODE_AZRANGE:

            tpColor3ub(0, (unsigned char)timer, (unsigned char)timer, 1);

            glBegin(GL_LINES);
               glVertex2iv(mouse_mode->box[0]);
               glVertex2iv(mouse_mode->box[1]);
            glEnd();

            tpColor3ub(255, 0, 0, 1);

            glBegin(GL_LINE_LOOP);
               for (i=0; i<32; i++)
                  glVertex2d(mouse_mode->box[0][0] + circle[i][0]*7, mouse_mode->box[0][1] + circle[i][1]*7);
            glEnd();

            tpColor3ub(0, 255, 255, 1);

            glBegin(GL_LINE_LOOP);
               for (i=0; i<32; i++)
                  glVertex2d(mouse_mode->box[1][0] + circle[i][0]*7, mouse_mode->box[1][1] + circle[i][1]*7);
            glEnd();

            DrawString(10,50,mouse_mode->bear_sreadout,strlen(mouse_mode->bear_sreadout));
            DrawString(10,40,mouse_mode->bear_ereadout,strlen(mouse_mode->bear_ereadout));
            DrawString(10,30,mouse_mode->bear_rangebear,strlen(mouse_mode->bear_rangebear));
            break;

         default:
            break;
      }

   newtime = clock();
   cyclic_timer += (90.0f*(newtime-oldtime))/CLOCKS_PER_SEC;
   
   if (cyclic_timer > 360.0f)
      cyclic_timer -= 360.0f;
   
   if (toggle) {
      timer -= (196.0f*(newtime-oldtime))/CLOCKS_PER_SEC;

      if (timer < 128) {
         timer = 128;
         toggle = 0;
      }

   }
  
   else {
      timer += (196.0f*(newtime-oldtime))/CLOCKS_PER_SEC;

      if (timer > 255) {
         timer = 255;
         toggle = 1;
      }

   }

   oldtime = newtime;

   win.VIRTUALswapbuffers();
   redraw_flag = 0;
}


/* ****************************************************************
**************************************************************** */
void canvas::DrawString(int x, int y, char *string, int length) {

   string_type buffer;
   
   buffer.stringsize(length+1);
   buffer.stringncpy(string, length);
   buffer.string[length] = 0;
   buffer.recalc();

#ifdef WIN32
   win.WOGLprint(x, y, buffer.string);
#else
   win.XOGLprint(x, y, buffer.string);
#endif
}

dbl_llist_manager canvas_manager;
       
