

/* ***********************************************************************
                Naval Air Warfare Center / Aircraft Division              
                        Patuxent Naval Air Station                         
 ========================================================================

        File:          sw_sweg_iface.c                                       
                                                                        
        Classification: UNCLASSIFIED                                   
                                                                        
        Description:    This file contains the functions, methods, and
			global variables needed to identify various
			data items in SWEG.
			
        Written by:	Alex Harper
*********************************************************************** */

#ifndef WIN32
#include <unistd.h>
#else
#include <math.h>
#endif

#include "global.h"
#include "timer.h"

#include "vissweg.h"


/* ***************************************************************************
  This method parses the command line for class specific items
*************************************************************************** */
sweg_iface::sweg_iface() {

   SWEDAT_LOCATION = USE_SHMEM;
   ASSET_TYPE      = MFS;
   USER_ID         = 40;  // MFS
   rec_mail = send_mail = NULL;
   deathcheck = -1.0;
}


/* *****************************************************************************
***************************************************************************** */
int sweg_iface::find_template(int type, Asset_action_message **aam) {

   if (!get_template(send_mail, -type, aam)) {
      while (!req_template(send_mail, type,1,0))
         SLEEP(0.01);

      while (!get_template(send_mail, -type, aam));
   }

   return 1;
}


/* ***************************************************************************
  This method parses the command line for class specific items
*************************************************************************** */
void sweg_iface::parse_input(int argc, char **argv) {

   int i;

   if (scramnet_flag)
      SWEDAT_LOCATION = USE_SCRAMNET;
   else
      SWEDAT_LOCATION = USE_SHMEM;

   for (i=1; i<argc; i++) {
      if (*(argv[i]) != '-')
         continue;
	 
      switch (*(argv[i]+1)) {

         case 'a':
	    
            if (!strcmp(argv[i]+1, "asset_type")) {
               i++;
 
               if (i == argc)
                  break;
	       
               ASSET_TYPE = atoi(argv[i]);
               break;
            }

            break;
	    
         case 'm':
            interface_flag |= INTERFACE_FLAG_MINICREW;
            break;
	    	     
         case 'o':
            interface_flag |= INTERFACE_FLAG_OVERRIDE;
            break;

         case 'r':
            if (!strcmp(argv[i]+1, "ready")) {
               interface_flag |= INTERFACE_FLAG_READY;
               break;
            }
	     
            break;

         case 's':
            if (!strcmp(argv[i]+1, "shmem")) {
               SWEDAT_LOCATION = USE_SHMEM;
               break;
            }
	     
            if (!strcmp(argv[i]+1, "scramnet")) {
               SWEDAT_LOCATION = USE_SCRAMNET;
               break;
            }

            break;
	
         case 't':
            if (!strcmp(argv[i]+1, "tdb")) {
               i++;
               if (i >= argc)
                  break;

               tdbdatname.stringcpy(argv[i]);

               i++;
               if (i >= argc)
                  break;

               tdblisname.stringcpy(argv[i]);
               continue;
            }
	 
            break;

         case 'u':
            i++;
            if (i == argc)
               break;
	       
            USER_ID = atoi(argv[i]);
            break;

         default:
            break;
      }

   }

   iface::parse_input(argc, argv);   
}


/* ***************************************************************************
  This method creates the initial platform wrapper lists
*************************************************************************** */
basic_platform *sweg_iface::CreatePlatformList(double *machine_time) {

   int index;
   Platform *plat;
   double t;

   dest();

   index = sweg[2];

   while (index) {
      plat = (Platform *) &sweg[index];
      index = plat->next_plat_from_master;
      t = get_float(plat->game_time);
      AddPlatform(plat, &t, machine_time);
   }

   return (basic_platform *)head;
}


/* ***************************************************************************
*************************************************************************** */
Semantic_Correlation *sweg_iface::find_correlation(int code) {

   Semantic_Correlation *ptr;
   int scindex;

   // correlate flight object
   for (scindex = sweg[6]; scindex; scindex = ptr->Sh_Next) {
      ptr = (Semantic_Correlation *)&sweg[scindex];

      if (ptr->semcode == code)
         return ptr;
   }

   return NULL;
}


/* ***************************************************************************
*************************************************************************** */
Semantic_Correlation *sweg_iface::find_user_defined_correlation(int code) {

   Semantic_Correlation *ptr;
   int scindex;

   // correlate flight object
   for (scindex = sweg[6]; scindex; scindex = ptr->Sh_Next) {
      ptr = (Semantic_Correlation *)&sweg[scindex];

      if (ptr->idcode == code || ptr->extracode == code)
         return ptr;
   }

   return NULL;
}


/* ***************************************************************************
  This method adds a platform wrapper to the existing list
*************************************************************************** */
basic_platform *sweg_iface::AddPlatform(Platform *plat, double *sim_time, double *machine_time) {

   int scode, scode2;
   sweg_platform *p;
   Semantic_Correlation *ptr;

   if (plat->acc_flags == -2 || 
       ((plat->stat_flags / DEAD_PLATFORM) % 10) == 1) {
      sprintf(perror_buffer, "Warning: Attempting to \"add\" dead platform %d...\n", plat->glb_SWEG_plat_id);
      pprintf(perror_buffer);
      return NULL;
   }

   ptr = find_correlation(plat->plat_semantic_code);

   if (ptr) {
      scode = ptr->idcode;
      scode2 = ptr->extracode;
   }

   else
      scode = scode2 = 0;

   // build platform
   append(p = new sweg_platform, NULL);

   p->birth_time = *sim_time;
   p->plat = plat;
   p->type = scode ? scode : scode2;

   // Note: following 2 lines must be before p->init()      
   if (control_platform.plat && plat->glb_SWEG_plat_id == control_platform.plat->glb_SWEG_plat_id) {
      p->platform_flag |= PLATFORM_FLAG_CONTROL;
      p->platform_flag &= ~PLATFORM_FLAG_DEAD_RECOG;
   }

   if (!p->init(this)) {
      sprintf(perror_buffer, "Cant find visual for %d : %d\n", p->type, plat->plat_semantic_code);
      pprintf(perror_buffer);
   }

   return p;
}


/* ***************************************************************************
*************************************************************************** */
int sweg_iface::process_messages(Asset_action_message *mess, double *machine_time) {

   int i, j;
   Platform *plat;
   basic_platform *ptr;
   Scenario *scen;
   double t, multiple;
   sweg_weapon_type *str;
   Semantic_Correlation *sctr;
   System *system;
   Extra_message_information *emi;
   sweg_sensor_type *sstr;

   switch (mess->action_code) {

      case AAM_END_SCENARIO:
         return -1;      

      case AAM_LEAVES_THE_SCENARIO:
         i = ((Platevent *)&sweg[mess->extra_info1])->id_sender;  

         ptr = get_platform(i);

         if (ptr) {
            ptr->platform_flag |= PLATFORM_FLAG_DEAD;
            return 1;
         }

         sprintf(perror_buffer, "Warning: Unable to process death %d\n", i);
         pprintf(perror_buffer);
         return 0;
	    
      case AAM_GENERAL_BIRTH_ANNOUNCMENT:
      case AAM_GENERAL_BIRTH_ANNOUNCMENT1:   // optional G.B.A. - jimm - for interface purposes - same format

         if (!mess->extra_info1)
            return 0;

         plat = (Platform *)&sweg[mess->extra_info1];

         if (get_platform(plat->glb_SWEG_plat_id)) {
            sprintf(perror_buffer, "Warning: Attempted to \"add\" existing platform %d\n", plat->glb_SWEG_plat_id);
            pprintf(perror_buffer);
            return 1;
         }

         t = get_float(mess->posted_game_time);
         ptr = AddPlatform(plat, &t, machine_time);     

         if (!ptr) {
            sprintf(perror_buffer, "Warning: Unable to \"add\" platform %d\n", plat->glb_SWEG_plat_id);
            pprintf(perror_buffer);
            return 0;
         }

         scen = (Scenario *)&sweg[get_master()->scen_pars];
         t = get_float(scen->game_time);
         multiple = get_float(scen->mul_real_time);
         ptr->update(this, &t, &multiple, machine_time, scen->status == SWEG_PAUSE);
         return 1;

      case AAM_LETHAL_ENGAGE:
         if (!mess->extra_info1)
            return 0;

         system = ((System *)&sweg[mess->extra_info1]);
         sctr = find_correlation(system->semantic_code);

         if (!sctr) {
            sprintf(perror_buffer, "Warning: Unable to identify weapon semantic code %d...\n", system->semantic_code);
            pprintf(perror_buffer);
            return 0;
         }

         control_platform.weaponman.append(str = new sweg_weapon_type, NULL);
         str->weapon = system;
         str->semcode = sctr->idcode ? sctr->idcode : sctr->extracode;
         return 1;

      case AAM_MODE_CHANGE:
         if (!mess->extra_info2)
            return 0;

         emi = (Extra_message_information *)&sweg[mess->extra_info2];

         system = (System *)&sweg[emi->extra_info1];
         i = ((Platform *)&sweg[system->parent_plat])->glb_SWEG_plat_id;
         
         for (ptr = (basic_platform *)head; ptr  && ptr->id != i; ptr = (basic_platform *)ptr->next);

         if (!ptr) {
            sprintf(perror_buffer, "Warning: Unable to resolve platform %d\n", i);
            pprintf(perror_buffer);
            return 0;      
         }

         j = system->glb_sys_id;
         for (sstr = (sweg_sensor_type *)((sweg_platform *)ptr)->sensorman.head; sstr && sstr->id != j; sstr = (sweg_sensor_type *)sstr->next);

         if (!sstr) {
            sprintf(perror_buffer, "Warning: Unable to resolve platform:sensor %d:%d\n", i, j);
            pprintf(perror_buffer);
            return 0;      
         }

         sstr->mode = emi->extra_info2;
         ptr->platform_flag |= PLATFORM_FLAG_FORCED_UPDATE;
         return 1;

      default:
         return 0;
   }	    

}


/* ***************************************************************************
  This method handles the VISSWEG-SWEG mailbox
*************************************************************************** */
int sweg_iface::get_messages(double *machine_time) {

   Asset_action_message *mess;
   int index;

   if (!rec_mail)
      return 1;

   index=rec_mail->message_list;
   while (index) {
      mess = (Asset_action_message *)&sweg[index];
      index = mess->next_action_from_mailbox;

      if (mess->action_code > 0) {
         switch (process_messages(mess, machine_time)) {
            case -1:
               mess->action_code = -mess->action_code;
               return 0;

//            case 1:
            default:
               break;
         }

         mess->action_code = -mess->action_code;
      }
      
   }

   return 1;
}


/* ***************************************************************************
  This method handles the VISSWEG-SWEG mailbox
*************************************************************************** */
int sweg_iface::find_messages(int type, int *last_index, Asset_action_message **mess) {

   if (!rec_mail)
      return 1;

   if (!*last_index)
      *last_index = rec_mail->message_list;
   else {
      *mess = (Asset_action_message *)&sweg[*last_index];
      *last_index = (*mess)->next_action_from_mailbox;
   }

   for (; *last_index; *last_index = (*mess)->next_action_from_mailbox) {
      *mess = (Asset_action_message *)&sweg[*last_index];
   
      if ((*mess)->action_code == type)
         return 1;
   }

   return 0;
}


/* ***************************************************************************
  This method initializes the initial connection to SWEDAT
*************************************************************************** */
void sweg_iface::init() {

   Asset_header  *asset;         		// pointers to laboratory system/player
   Scenario      *scen;	// scenario parameters block(4)
   basic_platform *ptr;
   int index;
   double multiplier;
   int status;
   TIME_STRUCT t;
   double currtime, machine_time;
    
   iface::init();
   
   if (tdblisname.string[0] && tdbdatname.string[0]) {
      if (tdb_parser.extract_data(tdbdatname.string, tdblisname.string))
         tdb_parser.data_cleanup();
   }

   if (SWEDAT_LOCATION)
      map_sweg();
   else {
      set_scramnet_parms(0xdadda, 0, 1024*1024);
      map_bit3();
   }

   index = (get_master())->scen_pars;

   if (!index) {
      pprintf("ERROR: SWEG is not running... Aborting..\n");
      exit(0);
   }

   scen = (Scenario *)&sweg[index];

   while(scen->status != 1 && !(interface_flag & INTERFACE_FLAG_OVERRIDE)) {
#ifdef WIN32
      Sleep(1000);
#else
      sleep(1);
#endif
   }

   if (!get_asset(ASSET_TYPE,USER_ID,&asset)) {
      sprintf(perror_buffer, "Couldn't find asset %d:%d...  Exiting\n", ASSET_TYPE, USER_ID);
      pprintf(perror_buffer);
      exit(-1);
   }

   init_earth(scen);	// SWEGLIB Earth parameters
   init_earth(get_float(scen->lat_cen), get_float(scen->long_cen), get_float(scen->radius_of_earth));

   //  get mailboxes to and from SWEG
   control_platform.plat = NULL;
   if (ASSET_TYPE == SWEG_MS)
      send_mail = rec_mail = NULL;
   else {
	   
      if (!get_mailboxes(SWEG_MS, asset, &send_mail, &rec_mail)) {
         sprintf(perror_buffer, "Couldn't find mailboxes exiting\n");
         pprintf(perror_buffer);
         exit(-1);
      }

      for (index = asset->controlled_entity; index; index = ((Controlled_entity *)&sweg[index])->next_entry) {
         if (((Controlled_entity *)&sweg[index])->index_type == PLATFORM && ((Controlled_entity *)&sweg[index])->activity == CNTR_PLAT && ((Controlled_entity *)&sweg[index])->spp) {
            control_platform.plat = (Platform *)&sweg[((Controlled_entity *)&sweg[index])->spp];
            control_platform.id = control_platform.plat->glb_SWEG_plat_id;
            break;
         }

      }

   }

   // initialize graphics objects
   GETTIME(&t);
   machine_time = TIME2DOUBLE(t);

   CreatePlatformList(&machine_time);
   currtime = get_float(scen->game_time);
   multiplier = get_float(scen->mul_real_time);
   status = scen->status == SWEG_PAUSE;
   
   for (ptr=(basic_platform *)head; ptr; ptr=(basic_platform *)ptr->next)
      ptr->update(this, &currtime, &multiplier, &machine_time, status);
      
   if (!(interface_flag & INTERFACE_FLAG_OVERRIDE) && get_messages(&machine_time)) {

      //  tell sweg this asset is ready
      asset->status = SWEG_ASSET_READY;

      // Wait for sweg to finish initialization
      if (!(interface_flag & INTERFACE_FLAG_READY))
         pprintf("Waiting for start of simulation...\n");

      while (scen->status != SWEG_START_SCENARIO)
          get_messages(&machine_time);

      if (!(interface_flag & INTERFACE_FLAG_READY))
         pprintf("Connection completed...\n");
   }

}


/* ***************************************************************************
  This method clears out the dead platforms out of the wrapper lists
*************************************************************************** */
void sweg_iface::BurnDriftwood(double *machine_time) {

   sweg_platform *ptr;
   sweg_weapon_type *wtr, *xtr;
   sweg_sensor_type *str, *ttr;

   if (deathcheck > -0.5 && *machine_time < deathcheck)
      return;
  
   for (ptr = (sweg_platform *)head; ptr; ptr = (sweg_platform *)ptr->next)
      if (!(ptr->platform_flag & PLATFORM_FLAG_POSTMORTIUM) &&
          ((ptr->platform_flag & PLATFORM_FLAG_DEAD) ||
           ((ptr->plat->stat_flags / DEAD_PLATFORM) % 10) == 1 ||
           ptr->plat->acc_flags == -2)) {
         ptr->platform_flag |= PLATFORM_FLAG_DEAD | PLATFORM_FLAG_POSTMORTIUM;
         ptr->perform(this, PERFORM_PLATFORM_DEACTIVATE, NULL);
      }
   
      else {
         wtr = (sweg_weapon_type *)ptr->weaponman.head;
         while (wtr) {
            xtr = wtr;
            wtr = (sweg_weapon_type *)wtr->next;
            if ((xtr->weapon->status_flag / DEAD_SYSTEM) & 0x01) {
               ptr->weaponman.remove(xtr);
               delete xtr;
            }
         }

         str = (sweg_sensor_type *)ptr->sensorman.head;
         while (str) {
            ttr = str;
            str = (sweg_sensor_type *)str->next;
            if ((ttr->sensor_xmit->status_flag / DEAD_SYSTEM) & 0x01) {
               ptr->sensorman.remove(ttr);
               delete ttr;
            }
         }

      }

   deathcheck = *machine_time + 1.0;
}


/* ***************************************************************************
  This method updates the wrappers
*************************************************************************** */
void sweg_iface::update(double *machine_time) {

   basic_platform *ptr;
   Scenario *scen;
   double multiple;
   int status;
   double currtime;

   scen = (Scenario *)&sweg[sweg[3]];
   currtime = get_float(scen->game_time);
   multiple = get_float(scen->mul_real_time);
   status = scen->status == SWEG_PAUSE;

   if (!(interface_flag & INTERFACE_FLAG_MINICREW))
      for (ptr=(basic_platform *)head; ptr; ptr=(basic_platform *)ptr->next) {
         if (!(ptr->platform_flag & PLATFORM_FLAG_DEAD))
            ptr->update(this, &currtime, &multiple, machine_time, status);
      }

   else
      for (ptr=(basic_platform *)head; ptr; ptr=(basic_platform *)ptr->next)
         if (!(ptr->platform_flag & (PLATFORM_FLAG_CONTROL | PLATFORM_FLAG_EXTERNAL | PLATFORM_FLAG_DEAD)))
            ptr->update(this, &currtime, &multiple, machine_time, status);

   get_messages(machine_time);
   BurnDriftwood(machine_time);
}


/* ***************************************************************************
*************************************************************************** */
void sweg_iface::mcp(plat_stat_type *pc) {

   if (!(interface_flag & INTERFACE_FLAG_MINICREW) || !control_platform.plat)
      return;

   // in use
   if (control_platform.plat->acc_flags < 0)
      return;

   control_platform.update_platform(pc);
}


/* ***************************************************************************
*************************************************************************** */
void sweg_iface::label(dbl_llist_manager *label_manager) {

   basic_platform *ptr;

   for (ptr=(basic_platform *)head; ptr; ptr=(basic_platform *)ptr->next)
      if (!(ptr->platform_flag & PLATFORM_FLAG_CONTROL)) 
         ptr->label(label_manager);
}


/* ***************************************************************************
*************************************************************************** */
int sweg_iface::get_data(int type, void *data) {

   Scenario *scen;

   switch (type) {

      case IFACEDATA_SCENARIO_CENTER:
         ((iface_scenario_center_type *)data)->latlon[0] = CONVERT_ctrlat;
         ((iface_scenario_center_type *)data)->latlon[1] = CONVERT_ctrlon;
         return 1;

      case IFACEDATA_SCENARIO_BOUNDS:

         scen = (Scenario *)&(sweg[(get_master())->scen_pars]);

         ((iface_scenario_bounds_type *)data)->length = get_float(scen->max_x_coor)*get_float(scen->max_x_coor);
         ((iface_scenario_bounds_type *)data)->length = ((iface_scenario_bounds_type *)data)->length < 1.0 ? 1.0 : sqrt(((iface_scenario_bounds_type *)data)->length);

         ((iface_scenario_bounds_type *)data)->width = get_float(scen->max_y_coor)*get_float(scen->max_y_coor);
         ((iface_scenario_bounds_type *)data)->width = ((iface_scenario_bounds_type *)data)->width < 1.0 ? 1.0 : sqrt(((iface_scenario_bounds_type *)data)->width);
      
         return 1;

      default:
         break;
   }

   return iface::get_data(type, data);
}


/* ***************************************************************************
*************************************************************************** */
int sweg_iface::set_data(int type, void *data) {

   double dt, machine_time;
   TIME_STRUCT t;
   
   switch (type) {
      case PERFORM_ADD_PLATFORM:
         GETTIME(&t);
         machine_time = TIME2DOUBLE(t);
         dt = get_float(((Platform *)data)->game_time);
         AddPlatform((Platform *)data, &dt, &machine_time);
         return 1;
      
      default:
         break;
   }
   
   return iface::set_data(type, data);
}

